from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Mapping


def _non_empty_string(value: Any) -> str | None:
    if not isinstance(value, str):
        return None
    text = value.strip()
    if text == "":
        return None
    return text


def _content_blocks(raw_message: Mapping[str, Any]) -> list[Mapping[str, Any]]:
    blocks = list[Mapping[str, Any]]()
    content = raw_message.get("content")
    if not isinstance(content, list):
        return blocks

    for raw_block in content:
        if isinstance(raw_block, dict):
            blocks.append(raw_block)
    return blocks


def _text_from_content_blocks(blocks: list[Mapping[str, Any]]) -> str:
    parts = list[str]()
    for block in blocks:
        if block.get("type") != "text":
            continue
        text = block.get("text")
        if isinstance(text, str):
            parts.append(text)
    return "".join(parts).strip()


def _attachment_file_names(raw_message: Mapping[str, Any]) -> list[str]:
    names = list[str]()

    for key in ("attachments", "files"):
        items = raw_message.get(key)
        if not isinstance(items, list):
            continue

        for raw_item in items:
            if not isinstance(raw_item, dict):
                continue
            file_name = _non_empty_string(raw_item.get("file_name"))
            if file_name is not None:
                names.append(file_name)

    deduped = list[str]()
    seen = set[str]()
    for name in names:
        lookup = name.lower()
        if lookup in seen:
            continue
        seen.add(lookup)
        deduped.append(name)
    return deduped


def _message_created_at(
    *,
    raw_message: Mapping[str, Any],
    blocks: list[Mapping[str, Any]],
) -> str | None:
    created_at = _non_empty_string(raw_message.get("created_at"))
    if created_at is not None:
        return created_at

    for block in blocks:
        start_timestamp = _non_empty_string(block.get("start_timestamp"))
        if start_timestamp is not None:
            return start_timestamp
        stop_timestamp = _non_empty_string(block.get("stop_timestamp"))
        if stop_timestamp is not None:
            return stop_timestamp

    return None


@dataclass(frozen=True)
class AnthropicThreadMessage:
    message_id: str
    author_name: str
    created_at: str
    text: str


@dataclass(frozen=True)
class AnthropicThreadConversation:
    conversation_id: str
    name: str
    created_at: str
    updated_at: str
    messages: tuple[AnthropicThreadMessage, ...]


class AnthropicThreadAdapter:
    def __init__(
        self,
        *,
        human_author_name: str = "human",
        assistant_author_name: str = "assistant",
        unknown_author_name: str = "participant",
        include_empty_messages: bool = False,
    ):
        self._human_author_name = human_author_name
        self._assistant_author_name = assistant_author_name
        self._unknown_author_name = unknown_author_name
        self._include_empty_messages = include_empty_messages

    def _author_name_for_sender(self, *, sender: str, raw_sender: str | None) -> str:
        if sender == "human":
            return self._human_author_name
        if sender == "assistant":
            return self._assistant_author_name
        if raw_sender is not None:
            return raw_sender
        return self._unknown_author_name

    def _message_text(
        self,
        *,
        raw_message: Mapping[str, Any],
        blocks: list[Mapping[str, Any]],
    ) -> str:
        if len(blocks) > 0:
            text = _text_from_content_blocks(blocks)
            if text != "":
                return text

        if len(blocks) == 0:
            fallback = _non_empty_string(raw_message.get("text"))
            if fallback is not None:
                return fallback

        file_names = _attachment_file_names(raw_message)
        if len(file_names) == 1:
            return f"attached file {file_names[0]}"
        if len(file_names) > 1:
            return "attached files " + ", ".join(file_names)
        return ""

    def conversation_from_export(
        self,
        *,
        raw_conversation: Mapping[str, Any],
    ) -> AnthropicThreadConversation:
        conversation_id = _non_empty_string(raw_conversation.get("uuid"))
        if conversation_id is None:
            raise ValueError("conversation is missing uuid")

        conversation_name = (
            _non_empty_string(raw_conversation.get("name")) or conversation_id
        )
        conversation_created_at = (
            _non_empty_string(raw_conversation.get("created_at")) or ""
        )
        conversation_updated_at = (
            _non_empty_string(raw_conversation.get("updated_at"))
            or conversation_created_at
        )

        raw_messages = raw_conversation.get("chat_messages")
        if not isinstance(raw_messages, list):
            raise ValueError("conversation chat_messages must be an array")

        messages = list[AnthropicThreadMessage]()
        for index, raw_message in enumerate(raw_messages):
            if not isinstance(raw_message, dict):
                continue

            blocks = _content_blocks(raw_message)
            text = self._message_text(raw_message=raw_message, blocks=blocks)
            if text == "" and not self._include_empty_messages:
                continue

            raw_sender = _non_empty_string(raw_message.get("sender"))
            sender = raw_sender.lower() if raw_sender is not None else ""
            author_name = self._author_name_for_sender(
                sender=sender,
                raw_sender=raw_sender,
            )

            message_created_at = _message_created_at(
                raw_message=raw_message,
                blocks=blocks,
            )
            if message_created_at is None:
                message_created_at = (
                    conversation_updated_at
                    if conversation_updated_at != ""
                    else conversation_created_at
                )

            message_id = (
                _non_empty_string(raw_message.get("uuid"))
                or f"{conversation_id}:message:{index}"
            )
            messages.append(
                AnthropicThreadMessage(
                    message_id=message_id,
                    author_name=author_name,
                    created_at=message_created_at,
                    text=text,
                )
            )

        return AnthropicThreadConversation(
            conversation_id=conversation_id,
            name=conversation_name,
            created_at=conversation_created_at,
            updated_at=conversation_updated_at,
            messages=tuple(messages),
        )

    def to_thread_json(
        self,
        *,
        conversation: AnthropicThreadConversation,
    ) -> dict[str, Any]:
        member_names = list[str]()
        seen_members = set[str]()
        for message in conversation.messages:
            member_name = message.author_name.strip()
            if member_name == "":
                continue
            if member_name in seen_members:
                continue
            seen_members.add(member_name)
            member_names.append(member_name)

        if len(member_names) == 0:
            member_names = [
                self._human_author_name,
                self._assistant_author_name,
            ]

        member_items = [
            {"member": {"name": member_name}} for member_name in member_names
        ]
        message_items = [
            {
                "message": {
                    "id": message.message_id,
                    "text": message.text,
                    "created_at": message.created_at,
                    "author_name": message.author_name,
                }
            }
            for message in conversation.messages
        ]

        return {
            "thread": {
                "name": conversation.name,
                "properties": [
                    {"members": {"items": member_items}},
                    {
                        "messages": {
                            "external_thread_id": conversation.conversation_id,
                            "items": message_items,
                        }
                    },
                ],
            }
        }



import os
import sys

import pytest

from meshagent.anthropic.messages_adapter import AnthropicMessagesAdapter
from meshagent.anthropic.mcp import MCPConfig, MCPServer, MCPTool
from meshagent.agents.agent import AgentSessionContext
from meshagent.tools import Toolkit


def _import_real_anthropic_sdk():
    """Import the external `anthropic` SDK without shadowing.

    If `pytest` is run from inside `.../meshagent/`, Python may resolve
    `import anthropic` to the local `meshagent/anthropic` package directory.
    """

    cwd = os.getcwd()

    if os.path.isdir(os.path.join(cwd, "anthropic")):
        sys.path = [p for p in sys.path if p not in ("", cwd)]

    import importlib

    mod = importlib.import_module("anthropic")

    mod_file = getattr(mod, "__file__", "") or ""
    if mod_file.endswith("/meshagent/anthropic/__init__.py"):
        raise RuntimeError(
            "Imported local `meshagent/anthropic` instead of the Anthropic SDK. "
            "Run pytest from the repo root."
        )

    return mod


a = _import_real_anthropic_sdk()


class _DummyRoom:
    # Adapter won't touch room when no tools.
    pass


@pytest.mark.asyncio
async def test_live_anthropic_adapter_messages_create_if_key_set():
    api_key = os.getenv("ANTHROPIC_API_KEY")
    if not api_key:
        pytest.skip("ANTHROPIC_API_KEY not set")

    model = os.getenv("ANTHROPIC_TEST_MODEL", "claude-sonnet-4-5")

    client = a.AsyncAnthropic(api_key=api_key)
    adapter = AnthropicMessagesAdapter(model=model, client=client, max_tokens=64)

    ctx = AgentSessionContext(system_role=None)
    ctx.append_user_message("Say hello in one word.")

    text = await adapter.next(context=ctx, room=_DummyRoom(), toolkits=[])

    assert isinstance(text, str)
    assert len(text.strip()) > 0


@pytest.mark.asyncio
async def test_live_anthropic_adapter_streaming_if_key_set():
    api_key = os.getenv("ANTHROPIC_API_KEY")
    if not api_key:
        pytest.skip("ANTHROPIC_API_KEY not set")

    model = os.getenv("ANTHROPIC_TEST_MODEL", "claude-sonnet-4-5")

    client = a.AsyncAnthropic(api_key=api_key)
    adapter = AnthropicMessagesAdapter(model=model, client=client, max_tokens=64)

    ctx = AgentSessionContext(system_role=None)
    ctx.append_user_message("Count from 1 to 3.")

    seen_types: list[str] = []

    def handler(event: dict):
        if isinstance(event, dict) and "type" in event:
            seen_types.append(event["type"])

    text = await adapter.next(
        context=ctx,
        room=_DummyRoom(),
        toolkits=[],
        event_handler=handler,
    )

    assert isinstance(text, str)
    assert len(text.strip()) > 0
    # These are best-effort; event types depend on Anthropic SDK.
    assert len(seen_types) > 0


@pytest.mark.asyncio
async def test_live_anthropic_mcp_deepwiki_if_key_set():
    api_key = os.getenv("ANTHROPIC_API_KEY")
    if not api_key:
        pytest.skip("ANTHROPIC_API_KEY not set")

    model = os.getenv("ANTHROPIC_TEST_MODEL", "claude-sonnet-4-5")

    client = a.AsyncAnthropic(api_key=api_key)
    adapter = AnthropicMessagesAdapter(model=model, client=client, max_tokens=256)

    ctx = AgentSessionContext(system_role=None)
    ctx.append_user_message(
        "Use the DeepWiki MCP toolset and make at least one tool call. "
        "Then reply with a one-sentence summary of what you learned."
    )

    mcp_toolkit = Toolkit(
        name="mcp",
        tools=[
            MCPTool(
                config=MCPConfig(
                    mcp_servers=[
                        MCPServer(url="https://mcp.deepwiki.com/mcp", name="deepwiki")
                    ]
                )
            )
        ],
    )

    seen_mcp_blocks = False

    def handler(event: dict):
        nonlocal seen_mcp_blocks
        if not isinstance(event, dict):
            return

        # Adapter forwards Anthropic SDK stream events:
        # {"type": "content_block_start", "event": {...}}
        if event.get("type") == "content_block_start":
            payload = event.get("event") or {}
            content_block = payload.get("content_block") or {}
            if content_block.get("type") in {"mcp_tool_use", "mcp_tool_result"}:
                seen_mcp_blocks = True

    text = await adapter.next(
        context=ctx,
        room=_DummyRoom(),
        toolkits=[mcp_toolkit],
        event_handler=handler,
    )

    assert isinstance(text, str)
    assert len(text.strip()) > 0

    # This asserts the connector actually engaged (best-effort, but should be stable
    # for DeepWiki).
    assert seen_mcp_blocks


@pytest.mark.asyncio
async def test_live_anthropic_adapter_compaction_if_key_set():
    api_key = os.getenv("ANTHROPIC_API_KEY")
    if not api_key:
        pytest.skip("ANTHROPIC_API_KEY not set")

    model = os.getenv(
        "ANTHROPIC_COMPACTION_TEST_MODEL",
        os.getenv("ANTHROPIC_TEST_MODEL", "claude-sonnet-4-6"),
    )

    client = a.AsyncAnthropic(api_key=api_key)
    adapter = AnthropicMessagesAdapter(
        model=model,
        client=client,
        max_tokens=64,
        context_management="auto",
        compaction_threshold=50000,
    )

    ctx = AgentSessionContext(system_role=None)
    ctx.append_user_message(
        "Reply with the single word 'ready'. Then wait for the next message."
    )

    try:
        first = await adapter.next(context=ctx, room=_DummyRoom(), toolkits=[])
    except Exception as ex:
        message = str(ex)
        if "does not support the 'compact_20260112'" in message:
            pytest.skip(f"model {model} does not support compact_20260112")
        raise
    assert isinstance(first, str)

    ctx.append_user_message("Now reply with the single word 'done'.")
    second = await adapter.next(context=ctx, room=_DummyRoom(), toolkits=[])
    assert isinstance(second, str)
    assert len(second.strip()) > 0
    assert ctx.metadata.get("last_response_model") == model
    usage = ctx.metadata.get("last_response_usage")
    assert isinstance(usage, dict)
    assert int(usage.get("input_tokens", 0)) > 0



from __future__ import annotations

from meshagent.agents.agent import AgentSessionContext
from meshagent.api import RoomClient, RoomException, RemoteParticipant
from meshagent.tools import Toolkit, ToolContext, FunctionTool, BaseTool
from meshagent.api.messaging import (
    Content,
    LinkContent,
    FileContent,
    JsonContent,
    TextContent,
    EmptyContent,
    RawOutputsContent,
    ensure_content,
)
from meshagent.agents.adapter import ToolResponseAdapter, LLMAdapter

import json
from typing import Any, Optional, Callable, Literal
from collections.abc import AsyncIterable
import os
import logging
import re
import asyncio
import base64
from html_to_markdown import convert

from meshagent.anthropic.proxy import get_client, get_logging_httpx_client
from meshagent.anthropic.mcp import MCPTool as MCPConnectorTool
from pydantic import BaseModel

try:
    from anthropic import APIStatusError
except Exception:  # pragma: no cover
    APIStatusError = Exception  # type: ignore

logger = logging.getLogger("anthropic_agent")

_CONTEXT_MANAGEMENT_BETA = "context-management-2025-06-27"
_COMPACTION_BETA = "compact-2026-01-12"


def _is_html_mime_type(mime_type: str | None) -> bool:
    if not mime_type:
        return False
    normalized = mime_type.split(";")[0].strip().lower()
    return normalized in {"text/html", "application/xhtml+xml"}


def _decode_text(data: bytes) -> str:
    return data.decode("utf-8", errors="replace")


def _replace_non_matching(text: str, allowed_chars: str, replacement: str) -> str:
    pattern = rf"[^{allowed_chars}]"
    return re.sub(pattern, replacement, text)


def safe_tool_name(name: str) -> str:
    return _replace_non_matching(name, "a-zA-Z0-9_-", "_")


def _as_jsonable(obj: Any) -> Any:
    if isinstance(obj, dict):
        return obj
    return obj.model_dump(
        mode="json",
        exclude_none=True,
        exclude_unset=True,
    )


def _text_block(text: str) -> dict:
    return {"type": "text", "text": text}


async def _consume_streaming_tool_result(
    *, stream: AsyncIterable[Any], event_handler: Optional[Callable[[dict], None]]
) -> Content:
    has_last = False
    last_item: Any = None
    async for item in stream:
        if (
            has_last
            and isinstance(last_item, JsonContent)
            and event_handler is not None
        ):
            event_handler(last_item.json)
        last_item = item
        has_last = True

    if not has_last:
        return ensure_content(None)

    if isinstance(last_item, dict):
        last_type = last_item.get("type")
        if last_type in ("agent.event", "codex.event"):
            return ensure_content(None)

    return ensure_content(last_item)


class AnthropicMessagesChatContext(AgentSessionContext):
    @property
    def supports_images(self) -> bool:
        return True

    @property
    def supports_files(self) -> bool:
        return True

    def append_image_message(self, *, mime_type: str, data: bytes) -> dict:
        normalized_mime_type = mime_type.lower().strip()
        if normalized_mime_type == "image/jpg":
            normalized_mime_type = "image/jpeg"

        allowed_mime_types = {"image/jpeg", "image/png", "image/gif", "image/webp"}
        if normalized_mime_type not in allowed_mime_types:
            message = {
                "role": "user",
                "content": [
                    _text_block(
                        f"the user attached an image in unsupported format {normalized_mime_type}"
                    )
                ],
            }
            self.messages.append(message)
            return message

        message = {
            "role": "user",
            "content": [
                {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": normalized_mime_type,
                        "data": base64.b64encode(data).decode("utf-8"),
                    },
                }
            ],
        }
        self.messages.append(message)
        return message

    def append_file_message(
        self, *, filename: str, mime_type: str, data: bytes
    ) -> dict:
        normalized_mime_type = (mime_type or "application/octet-stream").lower().strip()

        if normalized_mime_type.startswith("image/"):
            return self.append_image_message(
                mime_type=normalized_mime_type,
                data=data,
            )

        if normalized_mime_type == "application/pdf":
            message = {
                "role": "user",
                "content": [
                    {
                        "type": "document",
                        "title": filename,
                        "source": {
                            "type": "base64",
                            "media_type": normalized_mime_type,
                            "data": base64.b64encode(data).decode("utf-8"),
                        },
                    }
                ],
            }
            self.messages.append(message)
            return message

        if (
            normalized_mime_type.startswith("text/")
            or normalized_mime_type == "application/json"
            or normalized_mime_type == "application/xhtml+xml"
        ):
            if _is_html_mime_type(normalized_mime_type):
                text = convert(_decode_text(data))
            else:
                text = _decode_text(data)

            message = {
                "role": "user",
                "content": [
                    _text_block(
                        f"attached file {filename} ({normalized_mime_type}):\n{text}"
                    )
                ],
            }
            self.messages.append(message)
            return message

        message = {
            "role": "user",
            "content": [
                _text_block(
                    f"the user attached a file named {filename} with mime type {normalized_mime_type}"
                )
            ],
        }
        self.messages.append(message)
        return message


class MessagesToolBundle:
    def __init__(self, toolkits: list[Toolkit]):
        self._executors: dict[str, Toolkit] = {}
        self._safe_names: dict[str, str] = {}
        self._tools_by_safe_name: dict[str, FunctionTool] = {}

        tools: list[dict] = []

        for toolkit in toolkits:
            for v in toolkit.tools:
                if not isinstance(v, FunctionTool):
                    raise RoomException(f"unsupported tool type {type(v)}")

                original_name = v.name
                safe_name = safe_tool_name(original_name)

                if original_name in self._executors:
                    raise Exception(
                        f"duplicate in bundle '{original_name}', tool names must be unique."
                    )

                self._executors[original_name] = toolkit
                self._safe_names[safe_name] = original_name
                self._tools_by_safe_name[safe_name] = v

                schema = {**v.input_schema}
                if v.defs is not None:
                    schema["$defs"] = v.defs

                tools.append(
                    {
                        "name": safe_name,
                        "description": v.description,
                        "input_schema": schema,
                    }
                )

        self._tools = tools or None

    def to_json(self) -> list[dict] | None:
        return None if self._tools is None else self._tools.copy()

    def get_tool(self, safe_name: str) -> FunctionTool | None:
        return self._tools_by_safe_name.get(safe_name)

    async def execute(
        self, *, context: ToolContext, tool_use: dict
    ) -> Content | AsyncIterable[Any]:
        safe_name = tool_use.get("name")
        if safe_name not in self._safe_names:
            raise RoomException(
                f"Invalid tool name {safe_name}, check the name of the tool"
            )

        name = self._safe_names[safe_name]
        if name not in self._executors:
            raise Exception(f"Unregistered tool name {name}")

        arguments = tool_use.get("input") or {}
        proxy = self._executors[name]
        result = await proxy.execute(
            context=context,
            name=name,
            input=JsonContent(json=arguments),
        )
        return result


class AnthropicMessagesToolResponseAdapter(ToolResponseAdapter):
    async def to_plain_text(self, *, room: RoomClient, response: Content) -> str:
        if isinstance(response, LinkContent):
            return json.dumps({"name": response.name, "url": response.url})
        if isinstance(response, JsonContent):
            return json.dumps(response.json)
        if isinstance(response, TextContent):
            return response.text
        if isinstance(response, FileContent):
            return response.name
        if isinstance(response, EmptyContent):
            return "ok"
        if isinstance(response, dict):
            return json.dumps(response)
        if isinstance(response, str):
            return response
        if response is None:
            return "ok"
        raise Exception("unexpected return type: {type}".format(type=type(response)))

    async def create_messages(
        self,
        *,
        context: AgentSessionContext,
        tool_call: Any,
        room: RoomClient,
        response: Content,
    ) -> list:
        tool_use = tool_call if isinstance(tool_call, dict) else _as_jsonable(tool_call)
        tool_use_id = tool_use.get("id")
        if tool_use_id is None:
            raise RoomException("anthropic tool_use block was missing an id")

        if isinstance(response, RawOutputsContent):
            # Allow advanced tools to return pre-built Anthropic blocks.
            return [{"role": "user", "content": response.outputs}]

        tool_result_content: list[dict]
        try:
            if isinstance(response, FileContent):
                mime_type = (response.mime_type or "").lower()

                if mime_type == "image/jpg":
                    mime_type = "image/jpeg"

                if mime_type.startswith("image/"):
                    allowed = {"image/jpeg", "image/png", "image/gif", "image/webp"}
                    if mime_type not in allowed:
                        output = f"{response.name} was returned as {response.mime_type}, which Anthropic does not accept as an image block"
                        tool_result_content = [_text_block(output)]
                    else:
                        tool_result_content = [
                            {
                                "type": "image",
                                "source": {
                                    "type": "base64",
                                    "media_type": mime_type,
                                    "data": base64.b64encode(response.data).decode(
                                        "utf-8"
                                    ),
                                },
                            }
                        ]

                elif mime_type == "application/pdf":
                    tool_result_content = [
                        {
                            "type": "document",
                            "title": response.name,
                            "source": {
                                "type": "base64",
                                "media_type": "application/pdf",
                                "data": base64.b64encode(response.data).decode("utf-8"),
                            },
                        }
                    ]

                elif (
                    mime_type.startswith("text/")
                    or mime_type == "application/json"
                    or mime_type == "application/xhtml+xml"
                ):
                    if _is_html_mime_type(mime_type):
                        text = convert(_decode_text(response.data))
                    else:
                        text = _decode_text(response.data)
                    tool_result_content = [_text_block(text)]

                else:
                    output = await self.to_plain_text(room=room, response=response)
                    tool_result_content = [_text_block(output)]

            else:
                output = await self.to_plain_text(room=room, response=response)
                tool_result_content = [_text_block(output)]

        except Exception as ex:
            logger.error("unable to process tool call results", exc_info=ex)
            tool_result_content = [_text_block(f"Error: {ex}")]

        message = {
            "role": "user",
            "content": [
                {
                    "type": "tool_result",
                    "tool_use_id": tool_use_id,
                    "content": tool_result_content,
                }
            ],
        }

        room.developer.log_nowait(
            type="llm.message",
            data={
                "context": context.id,
                "participant_id": room.local_participant.id,
                "participant_name": room.local_participant.get_attribute("name"),
                "message": message,
            },
        )

        return [message]


class AnthropicMessagesAdapter(LLMAdapter[dict]):
    def __init__(
        self,
        model: str = os.getenv("ANTHROPIC_MODEL", "claude-3-5-sonnet-latest"),
        max_tokens: int = int(os.getenv("ANTHROPIC_MAX_TOKENS", "8192")),
        client: Optional[Any] = None,
        message_options: Optional[dict] = None,
        provider: str = "anthropic",
        log_requests: bool = False,
        context_management: Literal["auto", "none"] = "none",
        compaction_threshold: int = 150000,
        compaction_pause_after: bool = False,
        compaction_instructions: Optional[str] = None,
    ):
        if context_management not in ("auto", "none"):
            raise ValueError("context_management must be one of 'auto' or 'none'")
        if compaction_threshold < 50000:
            raise ValueError(
                "compaction_threshold must be greater than or equal to 50000"
            )
        self._model = model
        self._max_tokens = max_tokens
        self._client = client
        self._message_options = message_options or {}
        self._provider = provider
        self._log_requests = log_requests
        self._context_management_mode = context_management
        self._compaction_threshold = compaction_threshold
        self._compaction_pause_after = compaction_pause_after
        self._compaction_instructions = compaction_instructions

    def default_model(self) -> str:
        return self._model

    def create_session(self) -> AgentSessionContext:
        return AnthropicMessagesChatContext(system_role=None)

    def get_anthropic_client(self, *, room: RoomClient) -> Any:
        if self._client is not None:
            return self._client
        http_client = get_logging_httpx_client() if self._log_requests else None
        return get_client(room=room, http_client=http_client)

    def _convert_messages(
        self, *, context: AgentSessionContext
    ) -> tuple[list[dict], Optional[str]]:
        system = context.get_system_instructions()

        def as_blocks(role: str, content: Any) -> dict:
            if isinstance(content, str):
                return {"role": role, "content": [_text_block(content)]}
            if isinstance(content, list):
                return {"role": role, "content": content}
            return {"role": role, "content": [_text_block(str(content))]}

        messages: list[dict] = []
        pending_tool_use_ids: set[str] = set()

        for m in context.messages:
            role = m.get("role")
            if role not in {"user", "assistant"}:
                continue

            msg = as_blocks(role, m.get("content"))

            # Anthropic requires that tool_result blocks appear in the *immediately next*
            # user message after an assistant tool_use.
            if pending_tool_use_ids:
                if role == "assistant":
                    # Drop any assistant chatter that appears between tool_use and tool_result.
                    logger.warning(
                        "dropping assistant message between tool_use and tool_result"
                    )
                    continue

                # role == user
                content_blocks = msg.get("content") or []
                tool_results = [
                    b
                    for b in content_blocks
                    if isinstance(b, dict) and b.get("type") == "tool_result"
                ]
                tool_result_ids = {
                    b.get("tool_use_id") for b in tool_results if b.get("tool_use_id")
                }

                if not pending_tool_use_ids.issubset(tool_result_ids):
                    # If we can't satisfy the ordering contract, it's better to fail early
                    # with a clear error than to send an invalid request.
                    raise RoomException(
                        "invalid transcript: tool_use blocks must be followed by a user message "
                        "containing tool_result blocks for all tool_use ids"
                    )

                pending_tool_use_ids.clear()

            # Track tool_use ids introduced by assistant messages.
            if role == "assistant":
                content_blocks = msg.get("content") or []
                for b in content_blocks:
                    if isinstance(b, dict) and b.get("type") == "tool_use":
                        tool_id = b.get("id")
                        if tool_id:
                            pending_tool_use_ids.add(tool_id)

            messages.append(msg)

        return messages, system

    def _messages_api(self, *, client: Any, request: dict) -> Any:
        # The MCP connector requires `client.beta.messages.*`.
        if request.get("betas") is not None:
            return client.beta.messages
        extra_headers = request.get("extra_headers")
        if isinstance(extra_headers, dict) and extra_headers.get("anthropic-beta"):
            return client.beta.messages
        return client.messages

    async def _create_with_optional_headers(self, *, client: Any, request: dict) -> Any:
        api = self._messages_api(client=client, request=request)
        try:
            return await api.create(**request)
        except TypeError:
            request = dict(request)
            request.pop("extra_headers", None)
            return await api.create(**request)

    def _ensure_beta(self, *, request: dict[str, Any], beta: str) -> None:
        betas = request.get("betas")
        if betas is None:
            request["betas"] = [beta]
            return
        if isinstance(betas, str):
            normalized_betas = [betas]
        else:
            normalized_betas = [*betas]
        if beta not in normalized_betas:
            normalized_betas.append(beta)
        request["betas"] = normalized_betas

    def _ensure_context_management_betas(self, *, request: dict[str, Any]) -> None:
        context_management = request.get("context_management")
        if context_management is None:
            return
        if not isinstance(context_management, dict):
            self._ensure_beta(request=request, beta=_CONTEXT_MANAGEMENT_BETA)
            return
        edits = context_management.get("edits")
        if edits is None:
            self._ensure_beta(request=request, beta=_CONTEXT_MANAGEMENT_BETA)
            return

        has_compaction_edit = False
        for edit in edits:
            if isinstance(edit, dict) and edit.get("type") == "compact_20260112":
                has_compaction_edit = True
                break

        if has_compaction_edit:
            self._ensure_beta(request=request, beta=_COMPACTION_BETA)
            return
        self._ensure_beta(request=request, beta=_CONTEXT_MANAGEMENT_BETA)

    def _build_compaction_edit(self) -> dict[str, Any]:
        edit: dict[str, Any] = {
            "type": "compact_20260112",
            "trigger": {"type": "input_tokens", "value": self._compaction_threshold},
            "pause_after_compaction": self._compaction_pause_after,
        }
        if self._compaction_instructions is not None:
            edit["instructions"] = self._compaction_instructions
        return edit

    @staticmethod
    def _supports_auto_compaction_model(*, model: str) -> bool:
        normalized_model = model.strip().lower()
        if normalized_model == "":
            return False

        direct_match = re.match(r"^claude-(sonnet|opus)-(.+)$", normalized_model)
        if direct_match is not None:
            version_parts = direct_match.group(2).split("-")
            if len(version_parts) == 0 or not version_parts[0].isdigit():
                return False

            major = int(version_parts[0])
            minor = 0
            if (
                len(version_parts) > 1
                and version_parts[1].isdigit()
                and len(version_parts[1]) <= 2
            ):
                minor = int(version_parts[1])
            return (major, minor) > (4, 5)

        legacy_match = re.match(
            r"^claude-(\d+)-(\d+)-(sonnet|opus)(?:-|$)",
            normalized_model,
        )
        if legacy_match is not None:
            major = int(legacy_match.group(1))
            minor = int(legacy_match.group(2))
            return (major, minor) > (4, 5)

        return False

    def _add_auto_compaction_context_management(
        self, *, request: dict[str, Any]
    ) -> None:
        if self._context_management_mode != "auto":
            return
        model = request.get("model")
        if not isinstance(model, str) or not self._supports_auto_compaction_model(
            model=model
        ):
            return

        compaction_edit = self._build_compaction_edit()
        context_management = request.get("context_management")

        if context_management is None:
            request["context_management"] = {"edits": [compaction_edit]}
            return
        if not isinstance(context_management, dict):
            raise ValueError("context_management must be an object")

        edits_value = context_management.get("edits")
        if edits_value is None:
            edits: list[Any] = []
        else:
            edits = [*edits_value]

        has_compaction_edit = False
        normalized_edits = list[Any]()
        for edit in edits:
            if isinstance(edit, dict) and edit.get("type") == "compact_20260112":
                has_compaction_edit = True
            normalized_edits.append(edit)
        if not has_compaction_edit:
            normalized_edits.append(compaction_edit)

        request["context_management"] = {
            **context_management,
            "edits": normalized_edits,
        }

    def _normalize_usage(self, usage: object) -> dict[str, Any] | None:
        if usage is None:
            return None
        if isinstance(usage, dict):
            return usage
        if isinstance(usage, BaseModel):
            try:
                return usage.model_dump(mode="json")
            except Exception:
                return None
        return None

    def _store_usage(
        self, *, context: AgentSessionContext, response: dict[str, Any], model: str
    ) -> None:
        usage = self._normalize_usage(response.get("usage"))
        if usage is not None:
            context.metadata["last_response_usage"] = usage
            context.metadata["last_response_model"] = model
        context_management = response.get("context_management")
        if isinstance(context_management, dict):
            context.metadata["last_context_management"] = context_management

    async def _stream_message(
        self,
        *,
        client: Any,
        request: dict,
        event_handler: Callable[[dict], None],
    ) -> Any:
        """Stream text deltas and return the final message.

        Uses the official Anthropic SDK streaming helper:

        ```py
        async with client.messages.stream(...) as stream:
            async for text in stream.text_stream:
                ...
        message = await stream.get_final_message()
        ```
        """

        api = self._messages_api(client=client, request=request)
        stream_mgr = api.stream(**request)

        async with stream_mgr as stream:
            async for event in stream:
                event_handler({"type": event.type, "event": _as_jsonable(event)})

        final_message = await stream.get_final_message()
        event_handler(
            {"type": "message.completed", "message": _as_jsonable(final_message)}
        )
        return final_message

    def _split_toolkits(
        self, *, toolkits: list[Toolkit]
    ) -> tuple[list[Toolkit], list[BaseTool]]:
        """Split toolkits into executable tools and request middleware tools."""

        executable_toolkits: list[Toolkit] = []
        middleware: list[BaseTool] = []

        for toolkit in toolkits:
            executable_tools: list[FunctionTool] = []

            for t in toolkit.tools:
                if isinstance(t, MCPConnectorTool):
                    middleware.append(t)
                elif isinstance(t, FunctionTool):
                    executable_tools.append(t)
                elif isinstance(t, BaseTool):
                    if hasattr(t, "apply") and callable(getattr(t, "apply")):
                        middleware.append(t)
                    # Non-executable tool types are ignored.
                    continue
                else:
                    raise RoomException(f"unsupported tool type {type(t)}")

            if executable_tools:
                executable_toolkits.append(
                    Toolkit(
                        name=toolkit.name,
                        title=getattr(toolkit, "title", None),
                        description=getattr(toolkit, "description", None),
                        thumbnail_url=getattr(toolkit, "thumbnail_url", None),
                        rules=getattr(toolkit, "rules", []),
                        tools=executable_tools,
                    )
                )

        return executable_toolkits, middleware

    def _apply_request_middleware(
        self, *, request: dict, middleware: list[BaseTool]
    ) -> dict:
        headers = request.get("extra_headers") or {}
        for m in middleware:
            apply = getattr(m, "apply", None)
            if callable(apply):
                try:
                    apply(request=request, headers=headers)
                except TypeError:
                    apply(request=request)
        request["extra_headers"] = headers or None
        return request

    async def next(
        self,
        *,
        context: AgentSessionContext,
        room: RoomClient,
        toolkits: list[Toolkit],
        output_schema: Optional[dict] = None,
        event_handler: Optional[Callable[[dict], None]] = None,
        model: Optional[str] = None,
        on_behalf_of: Optional[RemoteParticipant] = None,
        options: Optional[dict] = None,
    ) -> Any:
        del options

        if model is None:
            model = self.default_model()

        tool_adapter = AnthropicMessagesToolResponseAdapter()

        client = self.get_anthropic_client(room=room)

        validation_attempts = 0

        try:
            while True:
                executable_toolkits, middleware = self._split_toolkits(
                    toolkits=toolkits
                )
                tool_bundle = MessagesToolBundle(toolkits=executable_toolkits)

                messages, system = self._convert_messages(context=context)

                extra_headers = {}
                if on_behalf_of is not None:
                    extra_headers["Meshagent-On-Behalf-Of"] = (
                        on_behalf_of.get_attribute("name")
                    )

                message_options = dict(self._message_options or {})

                tools_list: list[dict] = tool_bundle.to_json() or []
                extra_tools = message_options.pop("tools", None)
                if isinstance(extra_tools, list):
                    tools_list.extend(extra_tools)

                request = {
                    "model": model,
                    "max_tokens": self._max_tokens,
                    "messages": messages,
                    "system": system,
                    "tools": tools_list,
                    "extra_headers": extra_headers or None,
                    **message_options,
                }

                if output_schema is not None:
                    request["output_config"] = {
                        "format": {"type": "json_schema", "schema": output_schema}
                    }

                request = self._apply_request_middleware(
                    request=request,
                    middleware=middleware,
                )
                self._add_auto_compaction_context_management(request=request)
                self._ensure_context_management_betas(request=request)

                # Normalize empty lists to None for Anthropic.
                if (
                    isinstance(request.get("tools"), list)
                    and len(request["tools"]) == 0
                ):
                    request["tools"] = None
                if (
                    isinstance(request.get("mcp_servers"), list)
                    and len(request["mcp_servers"]) == 0
                ):
                    request["mcp_servers"] = None
                if (
                    isinstance(request.get("betas"), list)
                    and len(request["betas"]) == 0
                ):
                    request["betas"] = None

                # remove None fields
                request = {k: v for k, v in request.items() if v is not None}

                logger.info("requesting response from anthropic with model: %s", model)

                if event_handler is not None:
                    final_message = await self._stream_message(
                        client=client,
                        request=request,
                        event_handler=event_handler,
                    )
                    response_dict = _as_jsonable(final_message)
                else:
                    response = await self._create_with_optional_headers(
                        client=client,
                        request=request,
                    )
                    response_dict = _as_jsonable(response)

                self._store_usage(context=context, response=response_dict, model=model)

                content_blocks = []
                raw_content = response_dict.get("content")
                if isinstance(raw_content, list):
                    content_blocks = raw_content

                tool_uses = [b for b in content_blocks if b.get("type") == "tool_use"]

                # Keep the assistant message in context.
                assistant_message = {"role": "assistant", "content": content_blocks}
                context.messages.append(assistant_message)

                if tool_uses:
                    tasks = []

                    async def do_tool(tool_use: dict) -> list[dict]:
                        tool_context = ToolContext(
                            room=room,
                            caller=room.local_participant,
                            on_behalf_of=on_behalf_of,
                            caller_context={"chat": context.to_json()},
                        )
                        try:
                            tool_response = await tool_bundle.execute(
                                context=tool_context,
                                tool_use=tool_use,
                            )
                            if isinstance(tool_response, AsyncIterable):
                                tool_response = await _consume_streaming_tool_result(
                                    stream=tool_response,
                                    event_handler=event_handler,
                                )
                            else:
                                tool_response = ensure_content(tool_response)
                            return await tool_adapter.create_messages(
                                context=context,
                                tool_call=tool_use,
                                room=room,
                                response=tool_response,
                            )
                        except Exception as ex:
                            logger.error(
                                f"error in tool call {json.dumps(tool_use)}:",
                                exc_info=ex,
                            )
                            tool_result_content = [_text_block(f"Error: {ex}")]
                            message = {
                                "role": "user",
                                "content": [
                                    {
                                        "type": "tool_result",
                                        "tool_use_id": tool_use.get("id"),
                                        "content": tool_result_content,
                                    }
                                ],
                            }
                            return [message]

                    for tool_use in tool_uses:
                        tasks.append(asyncio.create_task(do_tool(tool_use)))

                    results = await asyncio.gather(*tasks)

                    # Anthropic requires tool_result blocks for *all* tool_use ids to appear in the
                    # *immediately next* user message after the assistant tool_use message.
                    tool_result_blocks: list[dict] = []
                    trailing_messages: list[dict] = []

                    for msgs in results:
                        for msg in msgs:
                            if (
                                isinstance(msg, dict)
                                and msg.get("role") == "user"
                                and isinstance(msg.get("content"), list)
                                and all(
                                    isinstance(b, dict)
                                    and b.get("type") == "tool_result"
                                    for b in msg["content"]
                                )
                            ):
                                tool_result_blocks.extend(msg["content"])
                            else:
                                trailing_messages.append(msg)

                    if tool_result_blocks:
                        context.messages.append(
                            {"role": "user", "content": tool_result_blocks}
                        )

                    for msg in trailing_messages:
                        context.messages.append(msg)

                    continue

                # no tool calls; return final content
                text = "".join(
                    [
                        b.get("text", "")
                        for b in content_blocks
                        if b.get("type") == "text"
                    ]
                )

                if output_schema is None:
                    return text

                # Schema-mode: parse and validate JSON.
                validation_attempts += 1
                try:
                    parsed = json.loads(text)
                    return parsed
                except Exception as e:
                    if validation_attempts >= 3:
                        raise RoomException(
                            f"Invalid JSON response from Anthropic: {e}"
                        )
                    context.messages.append(
                        {
                            "role": "user",
                            "content": (
                                "The previous response did not match the required JSON schema. "
                                f"Error: {e}. Please try again and return only valid JSON."
                            ),
                        }
                    )

        except APIStatusError as e:
            raise RoomException(f"Error from Anthropic: {e}")



from __future__ import annotations

from typing import Literal, Optional

from pydantic import BaseModel

from meshagent.tools import Toolkit, ToolkitBuilder, ToolkitConfig

from .request_tool import AnthropicRequestTool


class WebSearchConfig(ToolkitConfig):
    name: Literal["web_search"] = "web_search"
    max_uses: Optional[int] = None
    allowed_domains: Optional[list[str]] = None
    blocked_domains: Optional[list[str]] = None
    user_location: Optional["WebSearchUserLocation"] = None
    betas: Optional[list[str]] = None


class WebSearchUserLocation(BaseModel):
    type: Literal["approximate"] = "approximate"
    city: Optional[str] = None
    region: Optional[str] = None
    country: Optional[str] = None
    timezone: Optional[str] = None


class WebSearchTool(AnthropicRequestTool):
    def __init__(self, *, config: Optional[WebSearchConfig] = None):
        if config is None:
            config = WebSearchConfig(name="web_search")
        super().__init__(name=config.name)
        self.config = config

    def apply(self, *, request: dict, headers: dict) -> None:
        if self.config.allowed_domains and self.config.blocked_domains:
            raise ValueError(
                "web_search cannot set both allowed_domains and blocked_domains"
            )

        tools = request.setdefault("tools", [])
        tool_def: dict[str, object] = {
            "type": "web_search_20250305",
            "name": self.config.name,
        }
        if self.config.max_uses is not None:
            tool_def["max_uses"] = self.config.max_uses
        if self.config.allowed_domains is not None:
            tool_def["allowed_domains"] = self.config.allowed_domains
        if self.config.blocked_domains is not None:
            tool_def["blocked_domains"] = self.config.blocked_domains
        if self.config.user_location is not None:
            tool_def["user_location"] = self.config.user_location.model_dump(
                mode="json", exclude_none=True
            )
        if not any(
            isinstance(t, dict)
            and t.get("type") == tool_def["type"]
            and t.get("name") == tool_def["name"]
            for t in tools
        ):
            tools.append(tool_def)

        self.apply_betas(headers=headers, betas=self.config.betas)


class WebSearchToolkitBuilder(ToolkitBuilder):
    def __init__(self):
        super().__init__(name="web_search", type=WebSearchConfig)

    async def make(self, *, room, model: str, config: WebSearchConfig) -> Toolkit:
        return Toolkit(name="web_search", tools=[WebSearchTool(config=config)])



import pytest
from pydantic import BaseModel

from meshagent.anthropic.openai_responses_stream_adapter import (
    AnthropicOpenAIResponsesStreamAdapter,
)
from meshagent.anthropic.messages_adapter import AnthropicMessagesAdapter
from meshagent.agents.agent import AgentSessionContext


class _Event(BaseModel):
    type: str
    index: int | None = None
    message: dict | None = None
    content_block: dict | None = None
    delta: dict | None = None


class _FinalMessage(BaseModel):
    id: str = "msg_1"
    usage: dict = {"input_tokens": 3, "output_tokens": 5}


class _FakeStream:
    def __init__(self, events: list[BaseModel], final: BaseModel):
        self._events = events
        self._final = final

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        return False

    def __aiter__(self):
        async def gen():
            for e in self._events:
                yield e

        return gen()

    async def get_final_message(self):
        return self._final


class _FakeMessages:
    def __init__(self, stream: _FakeStream):
        self._stream = stream

    def stream(self, **kwargs):
        return self._stream


class _FakeClient:
    def __init__(self, stream: _FakeStream):
        self.messages = _FakeMessages(stream)


@pytest.mark.asyncio
async def test_openai_responses_stream_emits_content_part_events():
    events = [
        _Event(type="message_start", message={"id": "msg_1", "model": "claude"}),
        _Event(type="content_block_start", index=0, content_block={"type": "text"}),
        _Event(
            type="content_block_delta",
            index=0,
            delta={"type": "text_delta", "text": "hi"},
        ),
        _Event(type="content_block_stop", index=0),
        _Event(type="message_stop"),
    ]

    stream = _FakeStream(
        events=events,
        final=_FinalMessage(usage={"input_tokens": 3, "output_tokens": 5}),
    )
    client = _FakeClient(stream)

    adapter = AnthropicOpenAIResponsesStreamAdapter(client=client)

    emitted: list[dict] = []

    def handler(e: dict):
        emitted.append(e)

    await adapter._stream_message(
        client=client,
        request={"model": "x", "max_tokens": 5, "messages": []},
        event_handler=handler,
    )

    types = [e["type"] for e in emitted]

    assert "response.created" in types
    assert "response.output_item.added" in types
    assert "response.content_part.added" in types
    assert "response.output_text.delta" in types
    assert "response.output_text.done" in types
    assert "response.content_part.done" in types
    assert "response.output_item.done" in types
    assert "response.completed" in types

    # Sanity-check completed response contains usage.
    completed = next(e for e in emitted if e["type"] == "response.completed")
    assert completed["response"]["usage"]["input_tokens"] == 3
    assert completed["response"]["usage"]["output_tokens"] == 5
    assert completed["response"]["usage"]["total_tokens"] == 8


@pytest.mark.asyncio
async def test_openai_responses_stream_next_forwards_options(monkeypatch):
    called: dict = {}

    async def _fake_next(
        self,
        *,
        context,
        room,
        toolkits,
        output_schema=None,
        event_handler=None,
        model=None,
        on_behalf_of=None,
        options=None,
    ):
        del self
        del context
        del room
        del toolkits
        del output_schema
        del event_handler
        del model
        del on_behalf_of
        called["options"] = options
        return "ok"

    monkeypatch.setattr(AnthropicMessagesAdapter, "next", _fake_next)

    adapter = AnthropicOpenAIResponsesStreamAdapter(client=object())

    result = await adapter.next(
        context=AgentSessionContext(system_role=None),
        room=object(),
        toolkits=[],
        options={"reasoning": {"effort": "none"}},
    )

    assert result == "ok"
    assert called["options"] == {"reasoning": {"effort": "none"}}



from meshagent.anthropic.mcp import MCPConfig, MCPServer, MCPTool, MCPToolset


def test_mcp_tool_apply_injects_servers_toolsets_and_beta():
    cfg = MCPConfig(
        mcp_servers=[
            MCPServer(url="https://mcp.example.com/sse", name="example-mcp"),
        ],
        toolsets=[MCPToolset(mcp_server_name="example-mcp")],
    )

    tool = MCPTool(config=cfg)
    request: dict = {"tools": []}
    headers: dict = {}
    tool.apply(request=request, headers=headers)

    assert headers["anthropic-beta"] == "mcp-client-2025-11-20"
    assert request["mcp_servers"] == [
        {
            "type": "url",
            "url": "https://mcp.example.com/sse",
            "name": "example-mcp",
        }
    ]

    assert request["tools"][0]["type"] == "mcp_toolset"
    assert request["tools"][0]["mcp_server_name"] == "example-mcp"


def test_mcp_tool_apply_dedupes_servers_by_name_and_preserves_existing():
    cfg = MCPConfig(
        mcp_servers=[
            MCPServer(url="https://mcp.example.com/sse", name="example-mcp"),
            MCPServer(url="https://mcp.other.com/sse", name="other"),
        ]
    )
    tool = MCPTool(config=cfg)

    request: dict = {
        "tools": [{"type": "tool", "name": "some_tool"}],
        "mcp_servers": [
            {
                "type": "url",
                "url": "https://old.example.com/sse",
                "name": "example-mcp",
                "authorization_token": "OLD",
            }
        ],
    }

    headers: dict = {"anthropic-beta": "some-other-beta"}

    tool.apply(request=request, headers=headers)

    # Keeps existing betas and appends MCP beta.
    betas = {b.strip() for b in headers["anthropic-beta"].split(",") if b.strip()}
    assert "some-other-beta" in betas
    assert "mcp-client-2025-11-20" in betas

    # Dedupes by name; cfg overwrites the existing server entry.
    by_name = {s["name"]: s for s in request["mcp_servers"]}
    assert set(by_name.keys()) == {"example-mcp", "other"}
    assert by_name["example-mcp"]["url"] == "https://mcp.example.com/sse"

    # If toolsets omitted, it creates one per server.
    mcp_toolsets = [t for t in request["tools"] if t.get("type") == "mcp_toolset"]
    assert {t["mcp_server_name"] for t in mcp_toolsets} == {"example-mcp", "other"}



from meshagent.anthropic.thread_adapter import AnthropicThreadAdapter


def test_conversation_from_export_uses_text_blocks_and_skips_thinking() -> None:
    adapter = AnthropicThreadAdapter()

    conversation = adapter.conversation_from_export(
        raw_conversation={
            "uuid": "conv-1",
            "name": "Example",
            "created_at": "2026-02-01T00:00:00Z",
            "updated_at": "2026-02-01T00:10:00Z",
            "chat_messages": [
                {
                    "uuid": "msg-1",
                    "sender": "assistant",
                    "text": "thinking + answer merged in export",
                    "created_at": "2026-02-01T00:00:01Z",
                    "content": [
                        {"type": "thinking", "thinking": "internal chain of thought"},
                        {"type": "text", "text": "final answer"},
                    ],
                }
            ],
        }
    )

    assert conversation.conversation_id == "conv-1"
    assert len(conversation.messages) == 1
    assert conversation.messages[0].author_name == "assistant"
    assert conversation.messages[0].text == "final answer"


def test_conversation_from_export_uses_fallback_for_attachment_only_message() -> None:
    adapter = AnthropicThreadAdapter()

    conversation = adapter.conversation_from_export(
        raw_conversation={
            "uuid": "conv-1",
            "name": "Attachments",
            "created_at": "2026-02-01T00:00:00Z",
            "updated_at": "2026-02-01T00:10:00Z",
            "chat_messages": [
                {
                    "uuid": "msg-1",
                    "sender": "human",
                    "text": "",
                    "created_at": "2026-02-01T00:00:01Z",
                    "content": [],
                    "attachments": [{"file_name": "notes.txt"}],
                    "files": [],
                }
            ],
        }
    )

    assert len(conversation.messages) == 1
    assert conversation.messages[0].text == "attached file notes.txt"
    assert conversation.messages[0].author_name == "human"


def test_to_thread_json_contains_members_and_external_thread_id() -> None:
    adapter = AnthropicThreadAdapter()
    conversation = adapter.conversation_from_export(
        raw_conversation={
            "uuid": "conv-1",
            "name": "Thread Name",
            "created_at": "2026-02-01T00:00:00Z",
            "updated_at": "2026-02-01T00:10:00Z",
            "chat_messages": [
                {
                    "uuid": "msg-1",
                    "sender": "human",
                    "text": "",
                    "created_at": "2026-02-01T00:00:01Z",
                    "content": [{"type": "text", "text": "hello"}],
                },
                {
                    "uuid": "msg-2",
                    "sender": "assistant",
                    "text": "",
                    "created_at": "2026-02-01T00:00:02Z",
                    "content": [{"type": "text", "text": "hi"}],
                },
            ],
        }
    )

    thread_json = adapter.to_thread_json(conversation=conversation)
    messages = thread_json["thread"]["properties"][1]["messages"]
    members = thread_json["thread"]["properties"][0]["members"]["items"]

    assert thread_json["thread"]["name"] == "Thread Name"
    assert messages["external_thread_id"] == "conv-1"
    assert [item["member"]["name"] for item in members] == ["human", "assistant"]
    assert messages["items"][0]["message"]["text"] == "hello"



import pytest

from meshagent.anthropic.messages_adapter import (
    AnthropicMessagesAdapter,
    _consume_streaming_tool_result,
)
from meshagent.agents.agent import AgentSessionContext
from meshagent.api.messaging import JsonContent, TextContent
from meshagent.tools import FunctionTool, Toolkit
from meshagent.api import RoomException


class _DummyParticipant:
    def __init__(self):
        self.id = "p1"

    def get_attribute(self, name: str):
        if name == "name":
            return "tester"
        return None


class _DummyRoom:
    def __init__(self):
        self.local_participant = _DummyParticipant()
        self.developer = self

    def log_nowait(self, **kwargs):
        del kwargs


class _AnyArgsTool(FunctionTool):
    def __init__(self, name: str):
        super().__init__(
            name=name,
            input_schema={"type": "object", "additionalProperties": True},
            description="test tool",
        )

    async def execute(self, context, **kwargs):
        return {"ok": True, "args": kwargs}


class _FakeAdapter(AnthropicMessagesAdapter):
    def __init__(self, responses: list[dict], **kwargs):
        super().__init__(client=object(), **kwargs)
        self._responses = responses
        self._idx = 0
        self.requests: list[dict] = []

    async def _create_with_optional_headers(self, *, client, request):
        if self._idx >= len(self._responses):
            raise AssertionError("unexpected extra request")
        self.requests.append(request)
        resp = self._responses[self._idx]
        self._idx += 1
        return resp


class _StreamingTool(FunctionTool):
    def __init__(self, name: str):
        super().__init__(
            name=name,
            input_schema={"type": "object", "additionalProperties": True},
            description="streaming test tool",
        )

    async def execute(self, context, **kwargs):
        del context
        del kwargs

        async def _run():
            yield JsonContent(json={"type": "agent.event", "headline": "working"})
            yield TextContent(text="tool-final")

        return _run()


def test_convert_messages_drops_assistant_between_tool_use_and_tool_result():
    ctx = AgentSessionContext(
        system_role=None,
        messages=[
            {"role": "user", "content": "hi"},
            {
                "role": "assistant",
                "content": [
                    {"type": "text", "text": "calling tool"},
                    {
                        "type": "tool_use",
                        "id": "toolu_1",
                        "name": "tool_a",
                        "input": {},
                    },
                ],
            },
            {"role": "assistant", "content": "stray assistant message"},
            {
                "role": "user",
                "content": [
                    {
                        "type": "tool_result",
                        "tool_use_id": "toolu_1",
                        "content": [{"type": "text", "text": "ok"}],
                    }
                ],
            },
        ],
    )

    adapter = AnthropicMessagesAdapter(client=object())
    msgs, _system = adapter._convert_messages(context=ctx)

    assert [m["role"] for m in msgs] == ["user", "assistant", "user"]
    assert msgs[1]["content"][1]["type"] == "tool_use"


def test_convert_messages_raises_if_tool_result_not_immediately_next():
    ctx = AgentSessionContext(
        system_role=None,
        messages=[
            {"role": "user", "content": "hi"},
            {
                "role": "assistant",
                "content": [
                    {
                        "type": "tool_use",
                        "id": "toolu_1",
                        "name": "tool_a",
                        "input": {},
                    },
                ],
            },
            {"role": "user", "content": "not a tool_result"},
        ],
    )

    adapter = AnthropicMessagesAdapter(client=object())

    with pytest.raises(RoomException):
        adapter._convert_messages(context=ctx)


@pytest.mark.asyncio
async def test_next_batches_multiple_tool_results_into_single_user_message():
    responses = [
        {
            "content": [
                {"type": "text", "text": "calling tools"},
                {"type": "tool_use", "id": "toolu_1", "name": "tool_a", "input": {}},
                {"type": "tool_use", "id": "toolu_2", "name": "tool_b", "input": {}},
            ]
        },
        {"content": [{"type": "text", "text": "done"}]},
    ]

    adapter = _FakeAdapter(responses=responses)
    ctx = AgentSessionContext(system_role=None)
    ctx.append_user_message("run tools")

    toolkit = Toolkit(
        name="test",
        tools=[_AnyArgsTool("tool_a"), _AnyArgsTool("tool_b")],
    )

    result = await adapter.next(
        context=ctx,
        room=_DummyRoom(),
        toolkits=[toolkit],
    )

    assert result == "done"

    # Expect: user -> assistant(tool_use) -> user(tool_results batched) -> assistant(final)
    assert [m["role"] for m in ctx.messages] == [
        "user",
        "assistant",
        "user",
        "assistant",
    ]

    tool_results_msg = ctx.messages[2]
    assert tool_results_msg["role"] == "user"
    assert len(tool_results_msg["content"]) == 2
    assert {b["tool_use_id"] for b in tool_results_msg["content"]} == {
        "toolu_1",
        "toolu_2",
    }


@pytest.mark.asyncio
async def test_next_uses_final_stream_item_as_tool_result() -> None:
    responses = [
        {
            "content": [
                {"type": "text", "text": "calling tool"},
                {"type": "tool_use", "id": "toolu_1", "name": "tool_a", "input": {}},
            ]
        },
        {"content": [{"type": "text", "text": "done"}]},
    ]

    adapter = _FakeAdapter(responses=responses)
    ctx = AgentSessionContext(system_role=None)
    ctx.append_user_message("run tool")

    toolkit = Toolkit(name="test", tools=[_StreamingTool("tool_a")])
    result = await adapter.next(
        context=ctx,
        room=_DummyRoom(),
        toolkits=[toolkit],
    )

    assert result == "done"
    assert [m["role"] for m in ctx.messages] == [
        "user",
        "assistant",
        "user",
        "assistant",
    ]
    tool_result_content = ctx.messages[2]["content"][0]["content"][0]
    assert tool_result_content["type"] == "text"
    assert tool_result_content["text"] == "tool-final"


def test_create_chat_context_supports_images_and_files() -> None:
    adapter = AnthropicMessagesAdapter(client=object())
    context = adapter.create_session()

    assert context.supports_images is True
    assert context.supports_files is True

    image_message = context.append_image_message(mime_type="image/png", data=b"png")
    assert image_message["content"][0]["type"] == "image"

    file_message = context.append_file_message(
        filename="file.pdf",
        mime_type="application/pdf",
        data=b"%PDF-1.7",
    )
    assert file_message["content"][0]["type"] == "document"


def test_constructor_rejects_invalid_context_management_mode() -> None:
    with pytest.raises(
        ValueError, match="context_management must be one of 'auto' or 'none'"
    ):
        AnthropicMessagesAdapter(client=object(), context_management="invalid")


def test_constructor_rejects_invalid_compaction_threshold() -> None:
    with pytest.raises(
        ValueError,
        match="compaction_threshold must be greater than or equal to 50000",
    ):
        AnthropicMessagesAdapter(client=object(), compaction_threshold=49999)


@pytest.mark.asyncio
async def test_next_adds_auto_compaction_request_fields() -> None:
    adapter = _FakeAdapter(
        responses=[{"content": [{"type": "text", "text": "ok"}]}],
        model="claude-sonnet-4-6",
        context_management="auto",
        compaction_threshold=50001,
    )
    context = AgentSessionContext(system_role=None)
    context.append_user_message("hello")

    result = await adapter.next(
        context=context,
        room=_DummyRoom(),
        toolkits=[],
    )

    assert result == "ok"
    assert len(adapter.requests) == 1
    request = adapter.requests[0]
    assert request["context_management"] == {
        "edits": [
            {
                "type": "compact_20260112",
                "trigger": {"type": "input_tokens", "value": 50001},
                "pause_after_compaction": False,
            }
        ]
    }
    assert "compact-2026-01-12" in request["betas"]


@pytest.mark.asyncio
async def test_next_preserves_existing_betas_when_adding_compaction_beta() -> None:
    adapter = _FakeAdapter(
        responses=[{"content": [{"type": "text", "text": "ok"}]}],
        model="claude-opus-4-6",
        context_management="auto",
        message_options={"betas": ["mcp-client-2025-11-20"]},
    )
    context = AgentSessionContext(system_role=None)
    context.append_user_message("hello")

    result = await adapter.next(
        context=context,
        room=_DummyRoom(),
        toolkits=[],
    )

    assert result == "ok"
    assert len(adapter.requests) == 1
    request = adapter.requests[0]
    assert "mcp-client-2025-11-20" in request["betas"]
    assert "compact-2026-01-12" in request["betas"]


@pytest.mark.asyncio
async def test_next_skips_auto_compaction_for_legacy_model_versions() -> None:
    adapter = _FakeAdapter(
        responses=[{"content": [{"type": "text", "text": "ok"}]}],
        model="claude-sonnet-4-5",
        context_management="auto",
    )
    context = AgentSessionContext(system_role=None)
    context.append_user_message("hello")

    result = await adapter.next(
        context=context,
        room=_DummyRoom(),
        toolkits=[],
    )

    assert result == "ok"
    assert len(adapter.requests) == 1
    request = adapter.requests[0]
    assert "context_management" not in request
    assert "betas" not in request


@pytest.mark.asyncio
async def test_next_uses_context_management_beta_for_non_compaction_edits() -> None:
    adapter = _FakeAdapter(
        responses=[{"content": [{"type": "text", "text": "ok"}]}],
        message_options={
            "context_management": {
                "edits": [{"type": "clear_tool_uses_20250919", "trigger": "auto"}]
            }
        },
    )
    context = AgentSessionContext(system_role=None)
    context.append_user_message("hello")

    result = await adapter.next(
        context=context,
        room=_DummyRoom(),
        toolkits=[],
    )

    assert result == "ok"
    assert len(adapter.requests) == 1
    request = adapter.requests[0]
    assert "context-management-2025-06-27" in request["betas"]


@pytest.mark.asyncio
async def test_next_stores_usage_metadata() -> None:
    adapter = _FakeAdapter(
        responses=[
            {
                "content": [{"type": "text", "text": "ok"}],
                "usage": {
                    "input_tokens": 10,
                    "output_tokens": 5,
                    "iterations": [
                        {
                            "type": "compaction",
                            "input_tokens": 10,
                            "output_tokens": 5,
                            "cache_creation_input_tokens": 0,
                            "cache_read_input_tokens": 0,
                        }
                    ],
                },
            }
        ]
    )
    context = AgentSessionContext(system_role=None)
    context.append_user_message("hello")

    result = await adapter.next(
        context=context,
        room=_DummyRoom(),
        toolkits=[],
    )

    assert result == "ok"
    assert context.metadata["last_response_model"] == adapter.default_model()
    assert context.metadata["last_response_usage"]["input_tokens"] == 10
    assert context.metadata["last_response_usage"]["output_tokens"] == 5


@pytest.mark.asyncio
async def test_next_accepts_options_keyword_argument() -> None:
    adapter = _FakeAdapter(
        responses=[{"content": [{"type": "text", "text": "ok"}]}],
    )
    context = AgentSessionContext(system_role=None)
    context.append_user_message("hello")

    result = await adapter.next(
        context=context,
        room=_DummyRoom(),
        toolkits=[],
        options={"reasoning": {"effort": "none"}},
    )

    assert result == "ok"
    assert len(adapter.requests) == 1
    request = adapter.requests[0]
    assert "reasoning" not in request


class _ToolItemStream:
    def __init__(self, *, items: list[object]):
        self._items = items

    def __aiter__(self):
        return self._run()

    async def _run(self):
        for item in self._items:
            yield item


@pytest.mark.asyncio
async def test_consume_streaming_tool_result_emits_intermediate_json_chunk_events():
    events: list[dict] = []
    result = await _consume_streaming_tool_result(
        stream=_ToolItemStream(
            items=[
                JsonContent(json={"type": "agent.event", "headline": "working"}),
                TextContent(text="done"),
            ]
        ),
        event_handler=events.append,
    )

    assert events == [{"type": "agent.event", "headline": "working"}]
    assert isinstance(result, TextContent)
    assert result.text == "done"


@pytest.mark.asyncio
async def test_consume_streaming_tool_result_uses_final_item_as_result():
    events: list[dict] = []
    result = await _consume_streaming_tool_result(
        stream=_ToolItemStream(
            items=[
                JsonContent(json={"progress": 1}),
                JsonContent(json={"ok": True}),
            ]
        ),
        event_handler=events.append,
    )

    assert events == [{"progress": 1}]
    assert isinstance(result, JsonContent)
    assert result.json == {"ok": True}
